"""PowerPoint tools for video-to-GIF conversion and Google Slides upload."""

from .convert_pptx_videos_to_gifs import convert_pptx_videos_to_gifs
from .google_slides_upload import upload_and_merge_pptx
